!=========================================================================
! File GsHostProcess.gs
!
! Copyright (C) GemTalk Systems 2013-2016.  All Rights Reserved.
!
!=========================================================================

expectvalue %String
run
Globals at: #GsHostProcess ifAbsent:[
  Object _newKernelSubclass: 'GsHostProcess'
    subclassOf: Object
    instVarNames: #( cmd in out err childPid childStatus )
             " String, 3xGsSocket  , SmallInteger, SmallInteger "
    classVars: #()
    classInstVars: #()
    poolDictionaries: #()
    inDictionary: Globals
    constraints: #()
    options:  #( #instancesNonPersistent )
    reservedOop: 1213 .
  ^ ' created GsHostProcess'
].
^ 'exists'
%
run
| o |
(o := GsHostProcess asOop) == 155137 ifFalse:[ ^ o ].
^ true
%
set class GsHostProcess
removeallmethods
removeallclassmethods

category: 'Documentation'
classmethod:
comment
^ 'GsHostProcess represents a forked child process whose stdin, stdout,
and stderr are accessable via 3 instances of GsSocket. 

Instances may not be committed.

When an instance is garbage collected, and the child has not yet
been reaped,  the child process represented by that instance 
is killed with   kill -TERM  , and   waitpid   is called to
reap the child.

Lookup in the PATH environment variable is not performed. 
The argument to fork:  must specify a complete path to an executable 
or script.'
%

category: 'Instance creation'
classmethod: 
basicNew

"disallowed"

self shouldNotImplement: #basicNew
%

classmethod:
fork: commandLineString
" Create a new instance of the receiver, fork a child process,
 and exec the commandLineString in the child.
 Utf16 command lines are not yet supported by the _fork: primitive.
 Lookup in the PATH environment variable is not performed, 
 the commandLineString must specify a complete path to an executable 
 or script to be exec'ed .

 Signals an error if the privilege NoPerformOnServer is true."

  | inst |
  (inst := self _basicNew) _fork: commandLineString .
  ^ inst
%


category: 'Private'
classmethod:
_basicNew

"creates an instance registered with VM for finalization of cData"

<primitive: 674>
^ self _primitiveFailed: #_basicNew
%

method:
_fork: commandLineString 

"Signals an error if the privilege NoPerformOnServer is true."
 
<primitive: 956>
commandLineString _validateClass: String .
self _primitiveFailed: #_fork: args: { commandLineString } 
%

category: 'Accessing'
method:
commandLine
  "After fork: , returns the command line that was used to
   exec the child process."
  ^ cmd
%

method:
stdin
  "After fork: , returns a non-blocking write-only GsSocket 
   representing the parent's end of a pipe (see   man 2 pipe )  
   connected to stdin of the child.  The child's end of the pipe
   is blocking and a read by the child on stdin will block until
   Smalltalk code writes sufficient data to this GsSocket 
   to satisfy the read."
  ^ in
%

method:
stdout 
  "After fork: , returns a non-blocking read-only GsSocket 
  representing the parent's end of a pipe (see   man 2 pipe )  
  connected to stdout of the child.  If the GsSocket is not read
  from sufficiently to consume data, the child process may block 
  waiting to write more data to its stdout ."
  ^ out
%

method:
stderr
  "After fork: , returns a non-blocking read-only GsSocket 
  representing the parent's end of a pipe (see   man 2 pipe )  
  connected to stderr of the child.  If the GsSocket is not read
  from sufficiently to consume data, the child process may block 
  waiting to write more data to its stderr ."
  ^ err
%

method:
processId
  "After fork: , returns a SmallInteger , the operating system
   process id of the child."
  ^ childPid
%

category: 'Private'
method:
_waitChild

"calls waitpid() .
 Returns nil if child still running, otherwise returns a SmallInteger
 representing child status from waitpid() .

  result == 0 means child exited with exit code of 0 .

  ((result bitAnd: 16r100) ~~ 0)  means child exited due to a signal
     and  (result bitAnd: 16rFF) is the number of the signal .

  ((result bitAnd: 16r200) ~~ 0) means child was stopped by a signal
     and  (result bitAnd: 16rFF) is the number of the signal .

  otherwize non-zero result means child exited with that error status.
"
<primitive: 957>

self _primitiveFailed: #_waitChild
%

category: 'Status'
method:
childStatus
  "Returns nil if the child has not been reaped, 
   or returns a SmallInteger as follows

  result == 0 means child exited with exit code of 0 .

  ((result bitAnd: 16r100) ~~ 0)  means child exited due to a signal
     and  (result bitAnd: 16rFF) is the number of the signal .

  ((result bitAnd: 16r200) ~~ 0) means child was stopped by a signal 
     and  (result bitAnd: 16rFF) is the number of the signal .

  otherwize non-zero result means child exited with that error status."
 
  | s |
  (s := childStatus) ifNil:[ 
    s := self _waitChild  .
    s ifNotNil:[
      (s bitAnd: 16r200) == 0 ifTrue:[  "child exited"
        childStatus := s
        "in close . out close . err close ???" "close sockets to child"  
      ] 
    ].
  ].
  ^ s
%

method:
childHasExited
| s |
 s := self childStatus .
 s ifNil:[ ^ false "still running"].
 ^ (s  bitAnd: 16r200) == 0   "if not stopped with a signal it has exited"
%

