! ========================================================================
! GsLegacyExternalSession.gs
!
! Copyright (C) by GemTalk Systems 1991-2016.  All Rights Reserved
! ========================================================================

expectvalue %String
doit
(GsExternalSession subclass: 'GsLegacyExternalSession'
  instVarNames: #( gciInterface traversalBuffer)
  classVars: #()
  classInstVars: #()
  poolDictionaries: #()
  inDictionary: Globals
  options: #()) definition

%

! ------------------- Class comment for GsLegacyExternalSession
doit
GsLegacyExternalSession comment: 
'GsLegacyExternalSession is a variant of GsExternalSession which is used in
environments where the FFI GciLibrary cannot be used. It uses the GCI Interface
which is based on system primitives rather than FFI calls.

At this time, AIX is the only known environment where this is class must be used.
'.
true
%
doit
GsLegacyExternalSession category: 'External Sessions'.
true
%

! Remove existing behavior from GsLegacyExternalSession
doit
GsLegacyExternalSession removeAllMethods.
GsLegacyExternalSession class removeAllMethods.
true
%

! *** Subclass addition to the superclass.
category: 'Private'
set compile_env: 0
classmethod: GsExternalSession
sessionClass

	^(System gemVersionAt: 'osName') = 'AIX' 
		ifTrue: [GsLegacyExternalSession]
		ifFalse: [GsExternalSession]
%

! *** Override of the superclass' method (to be able to choose the suitable subclass) ***
category: 'Instance Creation'
set compile_env: 0
classmethod: GsExternalSession
new

	^(self sessionClass basicNew)
		initialize;
		yourself
%

! ------------------- Instance methods for GsLegacyExternalSession
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
abort

	self _errorIfCallInProgress.
	self _gciInterface nbAbort.
	self waitForResult.
%
category: 'Private'
set compile_env: 0
method: GsLegacyExternalSession
clearStackFor: anError

	| contextOop |
	self _gciInterface sessionId == 0 ifTrue: [^self].
	(contextOop := anError context) == nil asOop ifTrue: [^self].
	self _errorIfCallInProgress.
	self _gciInterface clearStack: contextOop.
	self _signalIfError.
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
commit

	self _errorIfCallInProgress.
	^self _gciInterface commit.
%
category: 'Private'
set compile_env: 0
method: GsLegacyExternalSession
continue: contextOop replacingTopOfStackWithOop: valueOop
	"Continue execution in the external Gem following an exception,
	 replacing the top of the stack with the specified object.
	 It is an error to specify an oop not visible to the remote Gem."
 
	self _errorIfCallInProgress. 
	^self _gciInterface 
		continueWith: contextOop 
		replaceTopOfStackWith: valueOop 
		isSpecial: false 
		flags: 0. 
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
forceLogout 
 
	self hardBreak. 
	self _logout. 
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
forkString: aString 
 
	self _errorIfCallInProgress. 
	self _gciInterface nbRemoteExecute: aString. 
%
category: 'Private'
set compile_env: 0
method: GsLegacyExternalSession
getAndClearLastError 
 
	^self _gciInterface getAndClearLastError. 
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
hardBreak 
 
	self _gciInterface hardBreak. 
%
category: 'Private'
set compile_env: 0
method: GsLegacyExternalSession
initialize

	super initialize.
	gciInterface := GciInterface new.
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
isCallInProgress 
	"The following calls are OK during a nonblocking call: 
	GciCallInProgress 
	GciErr 
	GciGetSessionId 
	GciHardBreak 
	GciNbEnd 
	GciSetSessionId 
	GciShutdown 
	GciSoftBreak" 
 
	self lastError notNil ifTrue: [self error: self lastError]. 
	self _gciInterface sessionId == 0 ifTrue: [^false]. 
	self _gciInterface callInProgress == true ifTrue: [^true]. 
	self getAndClearLastError. 
	^false. 
%
category: 'Private'
set compile_env: 0
method: GsLegacyExternalSession
_isOnMyStone 
 
	^parameters gemStoneName = System stoneName. 
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
isResultAvailable

	| progress |
	self isCallInProgress ifFalse: [self error: 'no call in progress'].
	progress := self _gciInterface nbEndOop.
	progress ifNil: [self _signalIfError].
	^progress = 1.
%
category: 'Private'
set compile_env: 0
method: GsLegacyExternalSession
lastError 
 
	^self _gciInterface lastError. 
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
lastResult

	| result error1 error2 |
	self _gciInterface resultIsSpecial ifTrue: [^self _gciInterface lastResult].
	error1 := self _gciInterface getAndClearLastError.
	self traversalBuffer initTraversalBuffer.
	result := self _gciInterface
		traverseObjects:  { self _gciInterface lastResult } 
		buffer: traversalBuffer.
	(error2 := self _gciInterface getAndClearLastError) notNil ifTrue: [ | txt |
		txt := error2 message.
		(txt == nil or: [txt trimBlanks isEmpty]) 
			ifTrue: [(Exception _new: error2 number args:  {} ) signal]
			ifFalse:[self error: txt]
	].
	traversalBuffer currentObjectIsSpecial ifTrue: [^traversalBuffer objId].
	traversalBuffer currentObjectIsByte ifTrue: [^traversalBuffer getByteObject].
	^ { traversalBuffer objId }
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
login

	| result |
	self _errorIfCallInProgress.
	result := self _gciInterface 
		login: parameters
		execute: 'System session'.
	self _signalIfError.
	stoneSessionId := result.
	self _isOnMyStone ifTrue: [
		stoneSessionSerial := GsSession serialOfSession: stoneSessionId.
		gemProcessId := (System descriptionOfSession: stoneSessionId) at: 2.
	] ifFalse: [
		stoneSessionSerial := self executeString: 'GsSession currentSession serialNumber'.
		gemProcessId := self executeString: 'System gemVersionAt: #''processId'''.
	].
	self log: 'GsLegacyExternalSession login: ', self _describe .
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
logout

	self isCallInProgress ifTrue: [
		[
			self waitForResult.
		] on: Error do: [:ex | 
			ex return.
		].
	].
	self _logout.
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
send: selector to: remoteOop withArguments: someValues

  | args |
  (someValues == nil or: [someValues isEmpty]) ifTrue: [ 
     args := #( )  
  ] ifFalse: [ 
    args := { } .
    someValues do: [:each | args add: each asOop ; add: each isSpecial ].
  ].
  self _gciInterface remotePerform: remoteOop selector: selector args: args.
  self _signalIfError.
  self _gciInterface resultIsSpecial ifTrue: [^self _gciInterface lastResult].
  ^ { self _gciInterface lastResult }
%
category: 'Accessors'
set compile_env: 0
method: GsLegacyExternalSession
sessionId

	^self _gciInterface sessionId.
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
softBreak 
 
	self _gciInterface softBreak. 
%
category: 'Public'
set compile_env: 0
method: GsLegacyExternalSession
traversalBuffer

	traversalBuffer ifNil: [
		(traversalBuffer := TraversalBuffer new: 8000)
			level: 1;
			initTraversalBuffer.
	].
	^traversalBuffer.
%
category: 'Private'
set compile_env: 0
method: GsLegacyExternalSession
_freeTraversalBuffer 
 
	traversalBuffer ifNil: [^self]. 
	traversalBuffer free.
	traversalBuffer := nil. 
%
category: 'Private'
set compile_env: 0
method: GsLegacyExternalSession
_gciInterface 
 
	^gciInterface
%
category: 'Private'
set compile_env: 0
method: GsLegacyExternalSession
_logout 
	 
	| descr | 
	descr := self _describe. 
	self _freeTraversalBuffer. 
	self _gciInterface logout. 
	self _waitForLogout. 
	self getAndClearLastError. 
	stoneSessionId := nil. 
	stoneSessionSerial := nil. 
	self log: 'GsLegacyExternalSession logout: ' , descr. 
%
category: 'Private'
set compile_env: 0
method: GsLegacyExternalSession
_signalIfError

	| error |
	(error := self _gciInterface getAndClearLastError) ifNil: [^self].
	GciLegacyError 
		signal: error 
		in: self.
%

