! =========================================================================
! Copyright (C) GemTalk Systems 1986-2016.  All Rights Reserved.
!
! $Id: ProfMonitorTree.gs 19139 2009-06-15 23:48:14Z lalmarod $
!
! Superclass Hierarchy:
!   ProfMonitorTree, ProfMonitor, Object.
!
!=========================================================================

expectvalue %String
run
| oldClass expectedInstVars newClass |
expectedInstVars := #(rawArray rootPME objRootPMEDict).
oldClass := Globals at: #ProfMonitorTree otherwise: nil.
(oldClass ~~ nil and: [oldClass instVarNames = expectedInstVars]) ifTrue: [
	^'Found expected class'.
].
newClass := ProfMonitor
  subclass: 'ProfMonitorTree'
  instVarNames: expectedInstVars
  classVars: #()
  classInstVars: #()
  poolDictionaries: { }
  inDictionary: Globals .
^(oldClass ifNil: ['Created new class:'] 
	ifNotNil: ['Replaced old class with:']) , '
' , newClass definition.

%

expectvalue %String
run
| curr n |
n := 0 .  "remove old methods for fix 44744"
(curr := ProfMonitorTree) classHistory size > 1 ifTrue:[
  curr classHistory do:[:cls |
    cls ~~ curr ifTrue:[
      cls removeAllMethods . cls class removeAllMethods .
      n := n + 1
    ].
  ].
].
^ 'found ', n asString , ' old versions'
%

removeallmethods ProfMonitorTree
removeallclassmethods ProfMonitorTree

category: 'For Documentation Installation only'
classmethod: ProfMonitorTree
installDocumentation

self comment: 'ProfMonitorTree is an enhancement of ProfMonitor, adding a tree based
display of the profiling results.' .

%

category: 'Private'
method: ProfMonitorTree
_methodTreeReportDownTo: tally
  "Return a report of the method call stack in tree form"

  | rpt elapsedTime |

  rpt := String new.

 " If there's no samples, bail"
  rootPME isNil ifTrue: [ ^rpt ].

  rpt add:'================'; lf ; 
      add: 'STACK SAMPLING TREE RESULTS'; lf .
  (endTime >= startTime) 
      ifTrue: [elapsedTime := endTime - startTime] 
      ifFalse: [elapsedTime := endTime + (16rffffffff - startTime)].
  rpt 
    add: 'elapsed CPU time:    '; 
      add: (elapsedTime) asString; add: ' ms';  lf ;
    add: (self _intervalString: interval ) ; lf ; lf .

  self _treeReportDownTo: tally on: rootPME to: rpt.

^ rpt
%

category: 'Private'
method: ProfMonitorTree
_objectCreationTreeReportDownTo: tally
  "report on object creation, in tree form."

  | rpt elapsedTime croots |

  rpt := String new.
  rpt add:'================'; lf ; 
      add: 'OBJECT CREATION TREE RESULTS'; lf .

  " If there's no samples, bail "
  (rootPME isNil or: [objRootPMEDict keys size < 1]) ifTrue: [ 
     rpt add: ' - - No results to report - -' ; lf.
     ^ rpt.
     ].

  (endTime >= startTime) 
    ifTrue: [elapsedTime := endTime - startTime] 
    ifFalse: [elapsedTime := endTime + (16rffffffff - startTime)].
  rpt 
     add: 'elapsed CPU time:    '; 
       add: (elapsedTime) asString; add: ' ms';  lf ;
        add: (self _intervalString: interval ) ; lf ; lf .

  croots := Array new.
  objRootPMEDict keysAndValuesDo: [ :cls :root |
    croots add: (Array with: cls with: root)].
  croots := croots 
    asSortedCollection: [:a :b | ((a at: 2) tally) > ((b at: 2) tally)].
  croots do: [:r | | cls root |
    cls := r at: 1.
    root := r at: 2.
    root tally < tally ifFalse: [
  		rpt add:'----------------'; lf ; add: cls name ; lf ; lf.
     		self _treeReportDownTo: tally on: root to: rpt ]].
^ rpt
%

category: 'Private'
method: ProfMonitorTree
gatherPMEsForTree

  | res rawArraySiz rawArrayPointer sampleStart sampleEnd reposCls |

  res := OrderedCollection new. 
  rawArray := self _readSampleFile .  
  (rawArraySiz := rawArray size) < 1 ifTrue:[ ^ self ].
  rawArrayPointer := 1 .
  rootPME := self readFirstPME.
  objRootPMEDict := IdentityKeyValueDictionary new.
  reposCls := SystemRepository.
  [ rawArrayPointer < rawArraySiz ] whileTrue:[
      (rawArray at: rawArrayPointer) == reposCls ifFalse:[
         self error: 'data order problem'
      ].
      sampleStart := rawArrayPointer + 2.   "skip count"
      sampleEnd := sampleStart + ((rawArray at: rawArrayPointer + 1) * 2) - 1.
      ((rawArray at: sampleStart) == true) 
      ifTrue: [
        "handle object trace report"
        sampleEnd := sampleEnd + 2.
        self tallyObjPME: (rawArray at: sampleStart + 1)
          position: sampleEnd 
          startIndex: sampleStart + 2
          endIndex: sampleEnd ]
      ifFalse: [
        "handle method trace report"
        self tallyPME: rootPME 
          position: sampleEnd 
          startIndex: sampleStart 
          endIndex: sampleEnd 
         ].
      rawArrayPointer := sampleEnd + 1.
  ].
%
category: 'Reporting'
method: ProfMonitorTree
gatherResults
  "See superclass comment"
  super gatherResults.
  self gatherPMEsForTree.
%
category: 'Private'
method: ProfMonitorTree
printablePercentFor: aFractionOrOtherNumber
  
  ^((aFractionOrOtherNumber asFloat * 100) 
    asStringUsingFormat: #(-6 1 false)) , '%'
%
category: 'Private'
method: ProfMonitorTree
printPME: aPME on: str total: theTotal
  
  str add: (self printablePercentFor: (aPME tally / theTotal)); space.
  str add: '('; add: aPME tally asString; add: ')'; space.

  str add: (aPME asStringWidth:20) ; lf
%
category: 'Private'
method: ProfMonitorTree
readFirstPME
  "trace consistes of SystemRepository, count, and (count*2) slots with method and class.
  The highest point in context is at end, create a PME based on that."

| index count |
index := 1.
(rawArray at: index) == SystemRepository 
  ifFalse: [self error: 'Cannot read first PME'].
count := (rawArray at: index + 1).
count _isSmallInteger ifFalse: [self error: 'corrupt prof monitor file'].

^ ProfMonitorEntry newForMethod: (rawArray at: (index + (count*2)))  
    receiverClass: (rawArray at: (index + (count*2) + 1)).
%

category: 'Private'
method: ProfMonitorTree
tallyPathPME: aPME position: index startIndex: startIndex endIndex: endIndex

  "Add a tally point for the PME's child, the one that matches the stack at 
  the given index.  Create child if does not exist. "

  | aMethod aClass matchingPME |
  aMethod := rawArray at: index - 1.
  aMethod class == GsNMethod ifFalse:[ aMethod := nil ].
  aClass := rawArray at: index .
  aPME ifNotNil:[
    aPME children do: [:aChildPME | 
      aChildPME cmethod == aMethod ifTrue:[ matchingPME := aChildPME ].
    ].
    aMethod ifNotNil:[ aClass ifNotNil:[
      matchingPME ifNil:[
        matchingPME := ProfMonitorEntry newForMethod: aMethod receiverClass: aClass. 
        aPME children add: matchingPME.
      ].
    ]].
  ].
  ^ matchingPME
%
category: 'Private'
method: ProfMonitorTree
tallyPME: aPME position: index startIndex: startIndex endIndex: endIndex
  " go backwards, top of stack is at end"

  | childPME idxMinus1 idxMinus2 mth |

  idxMinus1 := index - 1 .
  mth := rawArray at: idxMinus1 .
  mth ifNotNil:[ mth class == GsNMethod ifTrue:[ aPME ifNotNil:[
    mth == aPME cmethod ifTrue: [aPME tally: aPME tally + 1]
  ]]].

  idxMinus1 == startIndex ifTrue: [^self "end of path" ] .

  idxMinus2 := index - 2 .
  childPME := self tallyPathPME: aPME position: idxMinus2
            startIndex: startIndex endIndex: endIndex.

  self tallyPME: childPME 
      position: idxMinus2
      startIndex: startIndex endIndex: endIndex.
%

category: 'Private'
method: ProfMonitorTree
tallyObjPME: aClass position: index startIndex: startIndex endIndex: endIndex

  | aPME |
  (aPME := objRootPMEDict at: aClass otherwise: nil) ifNil: [
    aPME := self readFirstPME.
    objRootPMEDict at: aClass put: aPME ].
  ^ self tallyPME: aPME position: index startIndex: startIndex endIndex: endIndex
%

category: 'Private'
method: ProfMonitorTree
treePrintPME: aPME on: str tabs: tabs thisTab: myTab total: theTotal downTo: limit
  | children childrenTab |
  
  tabs do: [:tab | str add: tab].
  self printPME: aPME on: str total: theTotal.
  children := aPME children.
  children isEmpty ifFalse: [
      tabs addLast: myTab.
      children := children asSortedCollection: [:a :b | a tally > b tally].
      1 to: children size do: [:i | | child |
          child := children at: i.
          (child tally >= limit) ifTrue: [
            childrenTab := i < children size ifTrue: ['   |'] ifFalse: ['   '].
            self
              treePrintPME: child
              on: str
              tabs: tabs 
              thisTab: childrenTab
              total: theTotal
              downTo: limit
          ].
      ].
      tabs removeLast
    ]
%
category: 'Private'
method: ProfMonitorTree
_treeReportDownTo: tally on: root to: rpt

root children do: [:aPME |
  self 
    treePrintPME: aPME 
    on: rpt 
    tabs: OrderedCollection new  
    thisTab: '  ' 
    total: aPME tally
    downTo: tally
  ].
rpt lf; lf.
%
category: 'Reporting'
method: ProfMonitorTree
reportDownTo: tally
  "Formats and returns a string holding a report of the receiver's most 
   recent profile run. Only count at or above tally count are included in
   the report."

  | rpt |
  rpt := super reportDownTo: tally.
  rpt lf; lf.

  rpt add: (self _methodTreeReportDownTo: tally).
  traceObjCreation ifTrue: [
     rpt lf; lf.
     rpt add: (self _objectCreationTreeReportDownTo: tally).
     ].

^ rpt
%

