!=========================================================================
! Copyright (C) GemTalk Systems 2013-2016.  All Rights Reserved.
!=========================================================================

! Remove existing behavior from PPActionParser
removeallmethods PPActionParser
removeallclassmethods PPActionParser
! ------------------- Class methods for PPActionParser
category: 'instance creation'
set compile_env: 0
classmethod: PPActionParser
on: aParser block: aBlock
	^ (self on: aParser) setBlock: aBlock
%
! ------------------- Instance methods for PPActionParser
category: 'accessing'
set compile_env: 0
method: PPActionParser
block
	"Answer the action block of the receiver."

	^ block
%
category: 'parsing'
set compile_env: 0
method: PPActionParser
parseOn: aStream
	| element |
	^ (element := parser parseOn: aStream) isPetitFailure
		ifFalse: [ block value: element ]
		ifTrue: [ element ]
%
category: 'initialization'
set compile_env: 0
method: PPActionParser
setBlock: aBlockOrSymbol

	block := aBlockOrSymbol isSymbol
				ifFalse: [aBlockOrSymbol]
				ifTrue: [[:arg | arg perform: aBlockOrSymbol]]
%
expectValue %Boolean
doit
PPActionParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPAndParser
removeallmethods PPAndParser
removeallclassmethods PPAndParser
! ------------------- Class methods for PPAndParser
! ------------------- Instance methods for PPAndParser
category: 'operators'
set compile_env: 0
method: PPAndParser
and
	^ self
%
category: 'parsing'
set compile_env: 0
method: PPAndParser
parseOn: aStream
	| element position |
	position := aStream position.
	element := parser parseOn: aStream.
	aStream position: position.
	^ element
%
expectValue %Boolean
doit
PPAndParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPCharSetPredicate
removeallmethods PPCharSetPredicate
removeallclassmethods PPCharSetPredicate
! ------------------- Class methods for PPCharSetPredicate
category: 'instance creation'
set compile_env: 0
classmethod: PPCharSetPredicate
on: aBlock
	^ self basicNew initializeOn: aBlock
%
! ------------------- Instance methods for PPCharSetPredicate
category: 'initialization'
set compile_env: 0
method: PPCharSetPredicate
initializeOn: aBlock

	block := aBlock.
	classification := Array new: 255.
	1 to: classification size
		do: 
			[:index |
			classification at: index put: (block value: (Character withValue: index))]
%
category: 'evaluating'
set compile_env: 0
method: PPCharSetPredicate
value: aCharacter
	| index |
	index := aCharacter asInteger.
	index == 0
		ifTrue: [ ^ block value: aCharacter ].
	index > 255
		ifTrue: [ ^ block value: aCharacter ].
	^ classification at: index
%
expectValue %Boolean
doit
PPCharSetPredicate category: 'PetitParser-Tools'.
true
%

! Remove existing behavior from PPChoiceParser
removeallmethods PPChoiceParser
removeallclassmethods PPChoiceParser
! ------------------- Class methods for PPChoiceParser
! ------------------- Instance methods for PPChoiceParser
category: 'operators'
set compile_env: 0
method: PPChoiceParser
/ aRule 
	^ self copyWith: aRule
%
category: 'parsing'
set compile_env: 0
method: PPChoiceParser
parseOn: aStream
	"This is optimized code that avoids unnecessary block activations, do not change. When all choices fail, the last failure is answered."

	| element |
	1 to: parsers size do: [ :index |
		element := (parsers at: index)
			parseOn: aStream.
		element isPetitFailure
			ifFalse: [ ^ element ] ].
	^ element
%
expectValue %Boolean
doit
PPChoiceParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPCompositeParser
removeallmethods PPCompositeParser
removeallclassmethods PPCompositeParser
! ------------------- Class methods for PPCompositeParser
category: 'accessing'
set compile_env: 0
classmethod: PPCompositeParser
ignoredNames
	"Answer a collection of instance-variables that should not be automatically initialized with productions, but that are used internal to the composite parser."

	^ PPCompositeParser allInstVarNames
%
category: 'instance creation'
set compile_env: 0
classmethod: PPCompositeParser
new
	"Answer a new parser starting at the default start symbol."

	^ self newStartingAt: self startSymbol
%
category: 'instance creation'
set compile_env: 0
classmethod: PPCompositeParser
newStartingAt: aSymbol
	"Answer a new parser starting at aSymbol."

	^ self basicNew initializeStartingAt: aSymbol
%
category: 'parsing'
set compile_env: 0
classmethod: PPCompositeParser
parse: anObject
	^ self parse: anObject startingAt: self startSymbol
%
category: 'parsing'
set compile_env: 0
classmethod: PPCompositeParser
parse: anObject onError: aBlock
	^ self parse: anObject startingAt: self startSymbol onError: aBlock
%
category: 'parsing'
set compile_env: 0
classmethod: PPCompositeParser
parse: anObject startingAt: aSymbol
	^ (self newStartingAt: aSymbol) parse: anObject
%
category: 'parsing'
set compile_env: 0
classmethod: PPCompositeParser
parse: anObject startingAt: aSymbol onError: aBlock
	^ (self newStartingAt: aSymbol) parse: anObject onError: aBlock
%
category: 'accessing'
set compile_env: 0
classmethod: PPCompositeParser
startSymbol
	"Answer the method that represents the default start symbol."

	^ #start
%
! ------------------- Instance methods for PPCompositeParser
category: 'initialization'
set compile_env: 0
method: PPCompositeParser
initializeStartingAt: aSymbol
	| productionNames |
	self initialize.
	productionNames := self productionNames.
	parser := PPDelegateParser named: aSymbol.
	productionNames keysAndValuesDo: [ :key :value |
		self instVarAt: key put: (PPDelegateParser named: value) ].
	parser setParser: (self perform: aSymbol).
	productionNames keysAndValuesDo: [ :key :value |
		(self instVarAt: key) setParser: (self perform: value) ]
%
category: 'querying'
set compile_env: 0
method: PPCompositeParser
productionAt: aSymbol
	"Answer the production named aSymbol."
	
	^ self productionAt: aSymbol ifAbsent: [ nil ]
%
category: 'querying'
set compile_env: 0
method: PPCompositeParser
productionAt: aSymbol ifAbsent: aBlock
	"Answer the production named aSymbol, if there is no such production answer the result of evaluating aBlock."

	(self class ignoredNames includes: aSymbol asString)
		ifTrue: [^aBlock value].
	self class startSymbol = aSymbol ifTrue: [^parser].
	^self instVarAt: (self class allInstVarNames indexOf: aSymbol
				ifAbsent: [^aBlock value])
%
category: 'querying'
set compile_env: 0
method: PPCompositeParser
productionNames
	"Answer a dictionary of slot indexes and production names."
	
	| productionNames ignoredNames |
	productionNames := Dictionary new.
	ignoredNames := self class ignoredNames .
	self class allInstVarNames keysAndValuesDo: [ :key :value |
                value _isSymbol ifFalse:[ Error signal: 'expected a Symbol'].
		(ignoredNames includes: value )
			ifFalse: [ productionNames at: key put: value ] ].
	^ productionNames
%
category: 'accessing'
set compile_env: 0
method: PPCompositeParser
start
	"Answer the production to start this parser with."
	
	self subclassResponsibility: #start
%
expectValue %Boolean
doit
PPCompositeParser category: 'PetitParser-Tools'.
true
%

! Remove existing behavior from PPDelegateParser
removeallmethods PPDelegateParser
removeallclassmethods PPDelegateParser
! ------------------- Class methods for PPDelegateParser
category: 'instance creation'
set compile_env: 0
classmethod: PPDelegateParser
on: aParser
	^ self new setParser: aParser
%
! ------------------- Instance methods for PPDelegateParser
category: 'accessing'
set compile_env: 0
method: PPDelegateParser
children
	^ { parser }
%
category: 'parsing'
set compile_env: 0
method: PPDelegateParser
parseOn: aStream
	^ parser parseOn: aStream
%
category: 'initialization'
set compile_env: 0
method: PPDelegateParser
setParser: aParser
	parser := aParser
%
expectValue %Boolean
doit
PPDelegateParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPEndOfInputParser
removeallmethods PPEndOfInputParser
removeallclassmethods PPEndOfInputParser
! ------------------- Class methods for PPEndOfInputParser
! ------------------- Instance methods for PPEndOfInputParser
category: 'operators'
set compile_env: 0
method: PPEndOfInputParser
end
	^ self
%
category: 'parsing'
set compile_env: 0
method: PPEndOfInputParser
parseOn: aStream
	| position result |
	position := aStream position.
	result := parser parseOn: aStream.
	(result isPetitFailure or: [ aStream atEnd ])
		ifTrue: [ ^ result ].
	result := PPFailure
		message: 'end of input expected'
		at: aStream position.
	aStream position: position.
	^ result
%
expectValue %Boolean
doit
PPEndOfInputParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPEpsilonParser
removeallmethods PPEpsilonParser
removeallclassmethods PPEpsilonParser
! ------------------- Class methods for PPEpsilonParser
! ------------------- Instance methods for PPEpsilonParser
category: 'parsing'
set compile_env: 0
method: PPEpsilonParser
parseOn: aStream
	^ nil
%
expectValue %Boolean
doit
PPEpsilonParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPExpressionParser
removeallmethods PPExpressionParser
removeallclassmethods PPExpressionParser
! ------------------- Class methods for PPExpressionParser
! ------------------- Instance methods for PPExpressionParser
category: 'private'
set compile_env: 0
method: PPExpressionParser
build: aParser left: aChoiceParser

	^(aParser separatedBy: aChoiceParser) foldLeft: 
			[:a :op :b |
			op first
				value: a
				value: (op at: 2)
				value: b]
%
category: 'private'
set compile_env: 0
method: PPExpressionParser
build: aParser postfix: aChoiceParser

	^aParser , aChoiceParser star map: 
			[:term :ops |
			ops inject: term
				into: [:result :operator | operator first value: result value: (operator at: 2)]]
%
category: 'private'
set compile_env: 0
method: PPExpressionParser
build: aParser prefix: aChoiceParser

	^aChoiceParser star , aParser map: 
			[:ops :term |
			ops reverse inject: term
				into: [:result :operator | operator first value: (operator at: 2) value: result]]
%
category: 'private'
set compile_env: 0
method: PPExpressionParser
build: aParser right: aChoiceParser

	^(aParser separatedBy: aChoiceParser) foldRight: 
			[:a :op :b |
			op first
				value: a
				value: (op at: 2)
				value: b]
%
category: 'private'
set compile_env: 0
method: PPExpressionParser
buildOn: aParser

	^self buildSelectors inject: aParser
		into: 
			[:term :selector |
			| list |
			list := operators at: selector ifAbsent: [#()].
			list isEmpty
				ifTrue: [term]
				ifFalse: 
					[self
						perform: selector
						with: term
						with: (list size = 1
								ifTrue: 
									[list first first
										==> [:operator | { (list first at: 2) . operator} ]]
								ifFalse: 
									[list inject: PPChoiceParser new
										into: 
											[:choice :each |
											choice
												/ (each first ==> [:operator | { (each at: 2) . operator} ])]])]]
%
category: 'private'
set compile_env: 0
method: PPExpressionParser
buildSelectors
	^ #(build:prefix: build:postfix: build:right: build:left:)
%
category: 'specifying'
set compile_env: 0
method: PPExpressionParser
group: aOneArgumentBlock
	"Defines a priority group by evaluating aOneArgumentBlock."
	
	operators := Dictionary new.
	parser := [ 
		aOneArgumentBlock value: self.
	 	self buildOn: parser ]
			ensure: [ operators := nil ]
%
category: 'specifying'
set compile_env: 0
method: PPExpressionParser
left: aParser do: aThreeArgumentBlock
	"Define an operator aParser that is left-associative. Evaluate aThreeArgumentBlock with the first term, the operator, and the second term."
	
	self operator: #build:left: parser: aParser do: aThreeArgumentBlock
%
category: 'private'
set compile_env: 0
method: PPExpressionParser
operator: aSymbol parser: aParser do: aBlock
	parser isNil
		ifTrue: [ ^ self error: 'You did not specify a term when creating the receiver.' ].
	operators isNil
		ifTrue: [ ^ self error: 'Use #group: to define precedence groups in descending order.' ].
	(operators at: aSymbol ifAbsentPut: [ OrderedCollection new ])
		addLast: { aParser asParser . aBlock }
%
category: 'specifying'
set compile_env: 0
method: PPExpressionParser
postfix: aParser do: aTwoArgumentBlock
	"Define a postfix operator aParser. Evaluate aTwoArgumentBlock with the term and the operator."

	self operator: #build:postfix: parser: aParser do: aTwoArgumentBlock
%
category: 'specifying'
set compile_env: 0
method: PPExpressionParser
prefix: aParser do: aTwoArgumentBlock
	"Define a prefix operator aParser. Evaluate aTwoArgumentBlock with the operator and the term."

	self operator: #build:prefix: parser: aParser do: aTwoArgumentBlock
%
category: 'specifying'
set compile_env: 0
method: PPExpressionParser
right: aParser do: aThreeArgumentBlock
	"Define an operator aParser that is right-associative. Evaluate aThreeArgumentBlock with the first term, the operator, and the second term."
	
	self operator: #build:right: parser: aParser do: aThreeArgumentBlock
%
category: 'specifying'
set compile_env: 0
method: PPExpressionParser
term: aParser
	"Defines the initial term aParser of the receiver."
	
	parser isNil
		ifTrue: [ parser := aParser ]
		ifFalse: [ self error: 'Unable to redefine the term.' ]
%
expectValue %Boolean
doit
PPExpressionParser category: 'PetitParser-Tools'.
true
%

! Remove existing behavior from PPFailingParser
removeallmethods PPFailingParser
removeallclassmethods PPFailingParser
! ------------------- Class methods for PPFailingParser
category: 'instance creation'
set compile_env: 0
classmethod: PPFailingParser
message: aString
	^ self new setMessage: aString
%
! ------------------- Instance methods for PPFailingParser
category: 'accessing'
set compile_env: 0
method: PPFailingParser
message
	"Answer the error message of the receiving parser."

	^ message
%
category: 'parsing'
set compile_env: 0
method: PPFailingParser
parseOn: aStream
	^ PPFailure message: message at: aStream position
%
category: 'printing'
set compile_env: 0
method: PPFailingParser
printNameOn: aStream

	super printNameOn: aStream.
	aStream nextPutAll: ', '.
	message printOn: aStream
%
category: 'initialization'
set compile_env: 0
method: PPFailingParser
setMessage: aString
	message := aString
%
expectValue %Boolean
doit
PPFailingParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPFailure
removeallmethods PPFailure
removeallclassmethods PPFailure
! ------------------- Class methods for PPFailure
category: 'instance creation'
set compile_env: 0
classmethod: PPFailure
message: aString at: anInteger
	^ self basicNew initializeMessage: aString at: anInteger
%
! ------------------- Instance methods for PPFailure
category: 'initialization'
set compile_env: 0
method: PPFailure
initializeMessage: aString at: anInteger
	message := aString.
	position := anInteger
%
category: 'testing'
set compile_env: 0
method: PPFailure
isPetitFailure
  "I am the only class that should implement this method to return true."

  ^ true
%
category: 'accessing'
set compile_env: 0
method: PPFailure
message
	"Answer a human readable error message of this parse failure."
	
	^ message
%
category: 'accessing'
set compile_env: 0
method: PPFailure
position
	"Answer the position in the source string that caused this parse failure."

	^ position
%
category: 'printing'
set compile_env: 0
method: PPFailure
printOn: aStream

	aStream
		nextPutAll: self message;
		nextPutAll: ' at '.
	position printOn: aStream
%
expectValue %Boolean
doit
PPFailure category: 'PetitParser-Core'.
true
%

! Remove existing behavior from PPFlattenParser
removeallmethods PPFlattenParser
removeallclassmethods PPFlattenParser
! ------------------- Class methods for PPFlattenParser
! ------------------- Instance methods for PPFlattenParser
category: 'private'
set compile_env: 0
method: PPFlattenParser
create: aCollection start: aStartInteger stop: aStopInteger
	^ aCollection copyFrom: aStartInteger to: aStopInteger
%
category: 'parsing'
set compile_env: 0
method: PPFlattenParser
parseOn: aStream
	| start element stop |
	start := aStream position.
	element := parser parseOn: aStream.
	element isPetitFailure ifTrue: [
		aStream position: start.
		^ element ].
	stop := aStream position.
	^ self create: aStream collection start: start + 1 stop: stop
%
expectValue %Boolean
doit
PPFlattenParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPForwardingParser
removeallmethods PPForwardingParser
removeallclassmethods PPForwardingParser
! ------------------- Class methods for PPForwardingParser
category: 'instance creation'
set compile_env: 0
classmethod: PPForwardingParser
to: forwardee

	^self new forwardee: forwardee
%
! ------------------- Instance methods for PPForwardingParser
category: 'forwarding'
set compile_env: 0
method: PPForwardingParser
doesNotUnderstand: messageDescriptor
	"Forward them all"

	^forwardee
		perform: (messageDescriptor at: 1)
		env: 0
		withArguments: (messageDescriptor at: 2)
%
category: 'accessing'
set compile_env: 0
method: PPForwardingParser
forwardee: aParser

	forwardee := aParser
%
category: 'testing'
set compile_env: 0
method: PPForwardingParser
isPetitParser
	"Must override Object's implementation, or forward. Easier to override."

	^true
%
expectValue %Boolean
doit
PPForwardingParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPGreedyRepeatingParser
removeallmethods PPGreedyRepeatingParser
removeallclassmethods PPGreedyRepeatingParser
! ------------------- Class methods for PPGreedyRepeatingParser
! ------------------- Instance methods for PPGreedyRepeatingParser
category: 'parsing'
set compile_env: 0
method: PPGreedyRepeatingParser
parseOn: aStream
	| start element elements positions |
	start := aStream position.
	elements := OrderedCollection new.
	[ elements size < min ] whileTrue: [ 
		(element := parser parseOn: aStream) isPetitFailure ifTrue: [ 
			aStream position: start.
			^ element ].
		elements addLast: element ].
	positions := OrderedCollection with: aStream position.
	[ elements size < max and: [ (element := parser parseOn: aStream) isPetitFailure not ] ] whileTrue: [
		elements addLast: element.
		positions addLast: aStream position ].
	[ positions isEmpty ] whileFalse: [
		aStream position: positions last.
		element := limit parseOn: aStream.
		element isPetitFailure ifFalse: [
			aStream position: positions last.
			^ elements asArray ].
		elements isEmpty ifTrue: [
			aStream position: start.
			^ element ].
		elements removeLast.
		positions removeLast ].
	aStream position: start.
	^ PPFailure message: 'overflow' at: start
%
expectValue %Boolean
doit
PPGreedyRepeatingParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPLazyRepeatingParser
removeallmethods PPLazyRepeatingParser
removeallclassmethods PPLazyRepeatingParser
! ------------------- Class methods for PPLazyRepeatingParser
! ------------------- Instance methods for PPLazyRepeatingParser
category: 'parsing'
set compile_env: 0
method: PPLazyRepeatingParser
parseOn: aStream
	| start element elements |
	start := aStream position.
	elements := OrderedCollection new.
	[ elements size < min ] whileTrue: [
		(element := parser parseOn: aStream) isPetitFailure ifTrue: [
			aStream position: start.
			^ element ].
		elements addLast: element ].
	[ self matchesLimitOn: aStream ] whileFalse: [
		elements size < max ifFalse: [
			aStream position: start.
			^ PPFailure message: 'overflow' at: start ].
		element := parser parseOn: aStream.
		element isPetitFailure ifTrue: [
			aStream position: start.
			^ element ].
		elements addLast: element ].
	^ elements asArray
%
expectValue %Boolean
doit
PPLazyRepeatingParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPLimitedRepeatingParser
removeallmethods PPLimitedRepeatingParser
removeallclassmethods PPLimitedRepeatingParser
! ------------------- Class methods for PPLimitedRepeatingParser
category: 'instance creation'
set compile_env: 0
classmethod: PPLimitedRepeatingParser
on: aParser limit: aLimitParser
	^ (self on: aParser) setLimit: aLimitParser
%
! ------------------- Instance methods for PPLimitedRepeatingParser
category: 'accessing'
set compile_env: 0
method: PPLimitedRepeatingParser
children
	^ { parser . limit }
%
category: 'accessing'
set compile_env: 0
method: PPLimitedRepeatingParser
limit
	"Answer the parser that limits (or ends) this repetition."
	
	^ limit
%
category: 'private'
set compile_env: 0
method: PPLimitedRepeatingParser
matchesLimitOn: aStream
	| element position |
	position := aStream position.
	element := limit parseOn: aStream.
	aStream position: position.
	^ element isPetitFailure not
%
category: 'initialization'
set compile_env: 0
method: PPLimitedRepeatingParser
setLimit: aParser
	limit := aParser
%
expectValue %Boolean
doit
PPLimitedRepeatingParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPListParser
removeallmethods PPListParser
removeallclassmethods PPListParser
! ------------------- Class methods for PPListParser
category: 'instance creation'
set compile_env: 0
classmethod: PPListParser
with: aParser
	^ self withAll: { aParser }
%
category: 'instance creation'
set compile_env: 0
classmethod: PPListParser
with: aFirstParser with: aSecondParser
	^ self withAll: { aFirstParser . aSecondParser }
%
category: 'instance creation'
set compile_env: 0
classmethod: PPListParser
withAll: aCollection
	^ self basicNew setParsers: aCollection
%
! ------------------- Instance methods for PPListParser
category: 'accessing'
set compile_env: 0
method: PPListParser
children
	^ parsers
%
category: 'copying'
set compile_env: 0
method: PPListParser
copyWith: aParser
	^ self species withAll: (parsers copyWith: aParser)
%
category: 'initialization'
set compile_env: 0
method: PPListParser
initialize
	super initialize.
	self setParsers: #()
%
category: 'copying'
set compile_env: 0
method: PPListParser
postCopy
	super postCopy.
	parsers := parsers copy
%
category: 'initialization'
set compile_env: 0
method: PPListParser
setParsers: aCollection
	parsers := aCollection asArray
%
expectValue %Boolean
doit
PPListParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPLiteralObjectParser
removeallmethods PPLiteralObjectParser
removeallclassmethods PPLiteralObjectParser
! ------------------- Class methods for PPLiteralObjectParser
! ------------------- Instance methods for PPLiteralObjectParser
category: 'operators'
set compile_env: 0
method: PPLiteralObjectParser
caseInsensitive
  "Answer a parser that can parse the receiver case-insensitive."

  literal asUppercase = literal asLowercase
    ifTrue: [ ^ self ].
  ^ PPPredicateObjectParser
    on: [ :value | literal isEquivalent: value ]
    message: message
%
category: 'operators'
set compile_env: 0
method: PPLiteralObjectParser
negate
	^ (PPPredicateObjectParser expect: literal message: message) negate
%
category: 'parsing'
set compile_env: 0
method: PPLiteralObjectParser
parseOn: aStream
	^ (aStream atEnd not and: [ literal = aStream uncheckedPeek ])
		ifFalse: [ PPFailure message: message at: aStream position ]
		ifTrue: [ aStream next ]
%
expectValue %Boolean
doit
PPLiteralObjectParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPLiteralParser
removeallmethods PPLiteralParser
removeallclassmethods PPLiteralParser
! ------------------- Class methods for PPLiteralParser
category: 'instance creation'
set compile_env: 0
classmethod: PPLiteralParser
on: anObject
	^ self on: anObject message: anObject printString , ' expected'
%
category: 'instance creation'
set compile_env: 0
classmethod: PPLiteralParser
on: anObject message: aString
	^ self new initializeOn: anObject message: aString
%
! ------------------- Instance methods for PPLiteralParser
category: 'operators'
set compile_env: 0
method: PPLiteralParser
caseInsensitive
	"Answer a parser that can parse the receiver case-insensitive."
	
	self subclassResponsibility: #caseInsensitive
%
category: 'initialization'
set compile_env: 0
method: PPLiteralParser
initializeOn: anObject message: aString
	literal := anObject.
	message := aString
%
category: 'accessing'
set compile_env: 0
method: PPLiteralParser
literal
	"Answer the parsed literal."

	^ literal
%
category: 'accessing'
set compile_env: 0
method: PPLiteralParser
message
	"Answer the failure message."
	
	^ message
%
category: 'printing'
set compile_env: 0
method: PPLiteralParser
printNameOn: aStream

	super printNameOn: aStream.
	aStream nextPutAll: ', '.
	literal printOn: aStream
%
expectValue %Boolean
doit
PPLiteralParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPLiteralSequenceParser
removeallmethods PPLiteralSequenceParser
removeallclassmethods PPLiteralSequenceParser
! ------------------- Class methods for PPLiteralSequenceParser
! ------------------- Instance methods for PPLiteralSequenceParser
category: 'operators'
set compile_env: 0
method: PPLiteralSequenceParser
caseInsensitive
  "Answer a parser that can parse the receiver case-insensitive."

  literal asUppercase = literal asLowercase
    ifTrue: [ ^ self ].
  ^ PPPredicateSequenceParser
    on: [ :value | literal isEquivalent: value ]
    message: message
    size: size
%
category: 'initialization'
set compile_env: 0
method: PPLiteralSequenceParser
initializeOn: anObject message: aString
	super initializeOn: anObject message: aString.
	size := literal size
%
category: 'parsing'
set compile_env: 0
method: PPLiteralSequenceParser
parseOn: aStream
	| position result |
	position := aStream position.
	result := aStream next: size.
	literal = result ifTrue: [ ^ result ].
	aStream position: position.
	^ PPFailure message: message at: aStream position
%
category: 'accessing'
set compile_env: 0
method: PPLiteralSequenceParser
size
	"Answer the sequence size of the receiver."

	^ size
%
expectValue %Boolean
doit
PPLiteralSequenceParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPMemento
removeallmethods PPMemento
removeallclassmethods PPMemento
! ------------------- Class methods for PPMemento
category: 'instance creation'
set compile_env: 0
classmethod: PPMemento
new
	^ self basicNew initialize
%
! ------------------- Instance methods for PPMemento
category: 'accessing-readonly'
set compile_env: 0
method: PPMemento
count
	^ count
%
category: 'actions'
set compile_env: 0
method: PPMemento
increment
	count := count + 1
%
category: 'initialization'
set compile_env: 0
method: PPMemento
initialize
	count := 0
	
%
category: 'accessing'
set compile_env: 0
method: PPMemento
position
	^ position
%
category: 'accessing'
set compile_env: 0
method: PPMemento
position: anInteger
	position := anInteger
%
category: 'accessing'
set compile_env: 0
method: PPMemento
result
	^ result
%
category: 'accessing'
set compile_env: 0
method: PPMemento
result: anObject
	result := anObject
%
expectValue %Boolean
doit
PPMemento category: 'PetitParser-Core'.
true
%

! Remove existing behavior from PPMemoizedParser
removeallmethods PPMemoizedParser
removeallclassmethods PPMemoizedParser
! ------------------- Class methods for PPMemoizedParser
! ------------------- Instance methods for PPMemoizedParser
category: 'operators'
set compile_env: 0
method: PPMemoizedParser
memoized
	"Ther is no point in memoizing more than once."

	^ self
%
category: 'parsing'
set compile_env: 0
method: PPMemoizedParser
parseOn: aStream
	| memento |
	stream == aStream
		ifFalse: [ self reset: aStream ].
	memento := (buffer at: stream position + 1)
		ifNil: [ buffer at: stream position + 1 put: PPMemento new ].
	memento position isNil
		ifTrue: [
			memento result: (stream size - stream position + 2 < memento count
				ifTrue: [ PPFailure message: 'overflow' at: stream position ]
				ifFalse: [ memento increment. parser parseOn: stream ]).
			memento position: stream position ]
		ifFalse: [ stream position: memento position ].
	^ memento result
%
category: 'private'
set compile_env: 0
method: PPMemoizedParser
reset: aStream
	stream := aStream.
	buffer := Array new: aStream size + 1
%
expectValue %Boolean
doit
PPMemoizedParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPNotParser
removeallmethods PPNotParser
removeallclassmethods PPNotParser
! ------------------- Class methods for PPNotParser
! ------------------- Instance methods for PPNotParser
category: 'parsing'
set compile_env: 0
method: PPNotParser
parseOn: aStream
	| element position |
	position := aStream position.
	element := parser parseOn: aStream.
	aStream position: position.
	^ element isPetitFailure
		ifFalse: [ PPFailure message: '' at: aStream position ]
%
expectValue %Boolean
doit
PPNotParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPOptionalParser
removeallmethods PPOptionalParser
removeallclassmethods PPOptionalParser
! ------------------- Class methods for PPOptionalParser
! ------------------- Instance methods for PPOptionalParser
category: 'parsing'
set compile_env: 0
method: PPOptionalParser
parseOn: aStream
	| element |
	element := parser parseOn: aStream.
	^ element isPetitFailure ifFalse: [ element ]
%
expectValue %Boolean
doit
PPOptionalParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPParser
removeallmethods PPParser
removeallclassmethods PPParser
! ------------------- Class methods for PPParser
category: 'instance creation'
set compile_env: 0
classmethod: PPParser
named: aString
	^ self new name: aString
%
category: 'instance creation'
set compile_env: 0
classmethod: PPParser
new
	^ self basicNew initialize
%
! ------------------- Instance methods for PPParser
category: 'operators'
set compile_env: 0
method: PPParser
, aParser 
	"Answer a new parser that parses the receiver followed by aParser."

	^ PPSequenceParser with: self with: aParser
%
category: 'operators'
set compile_env: 0
method: PPParser
/ aParser 
	"Answer a new parser that parses the receiver, if the receiver fails try with aParser (ordered-choice)."
	
	^ PPChoiceParser with: self with: aParser
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
==> aBlock
	"Answer a new parser that performs aBlock as action handler on success."

	^ PPActionParser on: self block: aBlock
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
>=> aBlock
	"Answer a new parser that wraps the receiving parser with a two argument block. The first argument is the parsed stream, the second argument a continuation block on the delegate parser."

	^ PPWrappingParser on: self block: aBlock
%
category: 'operators'
set compile_env: 0
method: PPParser
and
	"Answer a new parser (logical and-predicate) that succeeds whenever the receiver does, but never consumes input."

	^ PPAndParser on: self
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
answer: anObject
	"Answer a new parser that always returns anObject from a successful parse."

	^ self ==> [ :nodes | anObject ]
%
category: 'converting'
set compile_env: 0
method: PPParser
asParser
	"Answer the receiving parser."
	
	^ self
%
category: 'accessing'
set compile_env: 0
method: PPParser
children
	"Answer a set of child parsers that could follow the receiver."

	^ #()
%
category: 'operators'
set compile_env: 0
method: PPParser
def: aParser
	"Redefine the receiver as the argument aParser. This method is useful when defining recursive parsers: instantiate a PPUnresolvedParser and later redefine it with another one."

	aParser name: self name.
	self become: (PPForwardingParser to: aParser).
	^self
%
category: 'operators-convenience'
set compile_env: 0
method: PPParser
delimitedBy: aParser
	"Answer a new parser that parses the receiver one or more times, separated and possibly ended by aParser."

	^((self separatedBy: aParser) , aParser optional) ==> 
			[:node |
			(node at: 2) isNil
				ifTrue: [node first]
				ifFalse: [node first copyWith: (node at: 2)]]
%
category: 'operators'
set compile_env: 0
method: PPParser
end
	"Answer a new parser that succeeds at the end of the input and return the result of the receiver."

	^ PPEndOfInputParser on: self
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
flatten
	"Answer a new parser that flattens the underlying collection."
	
	^ PPFlattenParser on: self
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
foldLeft: aBlock
	"Answer a new parser that that folds the result of the receiver from left-to-right into aBlock. The argument aBlock must take two or more arguments."
	
	| size args |
	size := aBlock numArgs.
	args := Array new: size.
	^ self ==> [ :nodes |
		args at: 1 put: nodes first.
		2 to: nodes size by: size - 1 do: [ :index |
			args
				replaceFrom: 2 to: size with: nodes startingAt: index;
				at: 1 put: (aBlock valueWithArguments: args) ].
		args first ]
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
foldRight: aBlock
	"Answer a new parser that that folds the result of the receiver from right-to-left into aBlock. The argument aBlock must take two or more arguments."

	| size args |
	size := aBlock numArgs.
	args := Array new: size.
	^ self ==> [ :nodes |
		args at: size put: nodes last.
		nodes size - size + 1 to: 1 by: 1 - size do: [ :index |
			args
				replaceFrom: 1 to: size - 1 with: nodes startingAt: index;
				at: size put: (aBlock valueWithArguments: args) ].
		args at: size ]
%
category: 'accessing-properties'
set compile_env: 0
method: PPParser
hasProperty: aKey
	"Test if the property aKey is present."
	
	^ properties notNil and: [ properties includesKey: aKey ]
%
category: 'initialization'
set compile_env: 0
method: PPParser
initialize
%
category: 'testing'
set compile_env: 0
method: PPParser
isPetitParser
	^ true
%
category: 'testing'
set compile_env: 0
method: PPParser
isUnresolved
	^ false
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
map: aBlock
	"Answer a new parser that works on the receiving sequence an passes in each element as a block argument."
	
	^ aBlock numArgs = 1
		ifTrue: [ self ==> aBlock ]
		ifFalse: [ self error: aBlock numArgs asString , ' arguments expected.' ]
%
category: 'parsing'
set compile_env: 0
method: PPParser
matches: anObject
	"Answer if anObject can be parsed by the receiver."
	
	^ (self parse: anObject) isPetitFailure not
%
category: 'parsing'
set compile_env: 0
method: PPParser
matchesIn: anObject
	"Search anObject repeatedly for the matches of the receiver. Answered an OrderedCollection of the matched parse-trees."

	| result |
	result := OrderedCollection new.
	self 
		matchesIn: anObject
		do: [ :each | result addLast: each ].
	^ result
%
category: 'parsing'
set compile_env: 0
method: PPParser
matchesIn: anObject do: aBlock
	"Search anObject repeatedly for the matches of the receiver. Evaluate aBlock for each match with the matched parse-tree as the argument. Make sure to always consume exactly one character with each step, to not miss any match."

	((self and ==> aBlock , #any asParser) / #any asParser) star parse: anObject
%
category: 'parsing'
set compile_env: 0
method: PPParser
matchesSkipIn: anObject
	"Search anObject repeatedly for the matches of the receiver. Answer an OrderedCollection of the matched parse-trees. Skip over matches."

	| result |
	result := OrderedCollection new.
	self 
		matchesSkipIn: anObject
		do: [ :each | result addLast: each ].
	^ result
%
category: 'parsing'
set compile_env: 0
method: PPParser
matchesSkipIn: anObject do: aBlock
	"Search anObject repeatedly for the matches of the receiver. Evaluate aBlock for each match with the matched parse-tree as the argument. Skip over matches."

	(self ==> aBlock / #any asParser) star parse: anObject
%
category: 'parsing'
set compile_env: 0
method: PPParser
matchingRangesIn: anObject
	"Search anObject repeatedly for the matches of the receiver. Answer an OrderedCollection of ranges of each match (index of first character to: index of last character)."
	
	| result |
	result := OrderedCollection new.
	self
		matchingRangesIn: anObject
		do: [ :value | result addLast: value ].
	^ result
%
category: 'parsing'
set compile_env: 0
method: PPParser
matchingRangesIn: anObject do: aBlock
	"Search anObject repeatedly for the matches of the receiver. Evaluate aBlock with the range of each match (index of first character to: index of last character)."
	
	[ :stream | stream position + 1 ] asParser , self , [ :stream | stream position ] asParser
		matchesIn: anObject
		do: [ :value | aBlock value: (value first to: value last) ]
%
category: 'parsing'
set compile_env: 0
method: PPParser
matchingSkipRangesIn: anObject
	"Search anObject repeatedly for the matches of the receiver. Skip over matches. Answer an OrderedCollection of ranges of each match (index of first character to: index of last character)."
	
	| result |
	result := OrderedCollection new.
	self
		matchingSkipRangesIn: anObject
		do: [ :value | result addLast: value ].
	^ result
%
category: 'parsing'
set compile_env: 0
method: PPParser
matchingSkipRangesIn: anObject do: aBlock
	"Search anObject repeatedly for the matches of the receiver. Skip over matches. Evaluate aBlock with the range of each match (index of first character to: index of last character)."
	
	[ :stream | stream position + 1 ] asParser , self , [ :stream | stream position ] asParser
		matchesSkipIn: anObject
		do: [ :value | aBlock value: (value first to: value last) ]
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
max: anInteger
	"Answer a new parser that parses the receiver at most anInteger times."
	
	^ self star setMax: anInteger
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
max: anInteger greedy: aParser
	"Answer a new parser that parses the receiver at most anInteger times until it reaches aParser. This is a greedy non-blind implementation. aParser is not consumed."
	
	^ (self starGreedy: aParser) setMax: anInteger
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
max: anInteger lazy: aParser
	"Answer a new parser that parses the receiver at most anInteger times until it reaches aParser. This is a lazy non-blind implementation. aParser is not consumed."
	
	^ (self starLazy: aParser) setMax: anInteger
%
category: 'operators'
set compile_env: 0
method: PPParser
memoized
	"Answer a new memoized parser, for refraining redundant computations. This ensures polynomial time O(n^4) for left-recursive grammars and O(n^3) for non left-recursive grammars in the worst case. Not necessary for most grammars that are carefully written and in O(n) anyway."
	
	^ PPMemoizedParser on: self
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
min: anInteger
	"Answer a new parser that parses the receiver at least anInteger times."
	
	^ self star setMin: anInteger
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
min: anInteger greedy: aParser
	"Answer a new parser that parses the receiver at least anInteger times until it reaches aParser. This is a greedy non-blind implementation. aParser is not consumed."
	
	^ (self starGreedy: aParser) setMin: anInteger
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
min: anInteger lazy: aParser
	"Answer a new parser that parses the receiver at least anInteger times until it reaches aParser. This is a lazy non-blind implementation. aParser is not consumed."
	
	^ (self starLazy: aParser) setMin: anInteger
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
min: aMinInteger max: aMaxInteger
	"Answer a new parser that parses the receiver at least aMinInteger and at most aMaxInteger times."
	
	^ self star setMin: aMinInteger; setMax: aMaxInteger
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
min: aMinInteger max: aMaxInteger greedy: aParser
	"Answer a new parser that parses the receiver at least aMinInteger and at most aMaxInteger times until it reaches aParser. This is a greedy non-blind implementation. aParser is not consumed."
	
	^ (self starGreedy: aParser) setMin: aMinInteger; setMax: aMaxInteger
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
min: aMinInteger max: aMaxInteger lazy: aParser
	"Answer a new parser that parses the receiver at least aMinInteger and at most aMaxInteger times until it reaches aParser. This is a greedy non-blind implementation. aParser is not consumed."
	
	^ (self starLazy: aParser) setMin: aMinInteger; setMax: aMaxInteger
%
category: 'accessing'
set compile_env: 0
method: PPParser
name
	"Answer the production name of the receiver."
	
	^ self propertyAt: #name ifAbsent: [ nil ]
%
category: 'accessing'
set compile_env: 0
method: PPParser
name: aString
	self propertyAt: #name put: aString
%
category: 'operators'
set compile_env: 0
method: PPParser
negate
	"Answer a new parser consumes any input token but the receiver."

	^(self not , #any asParser) ==> [:nodes | nodes at: 2]
%
category: 'operators'
set compile_env: 0
method: PPParser
not
	"Answer a new parser (logical not-predicate) that succeeds whenever the receiver fails, but never consumes input."

	^ PPNotParser on: self
%
category: 'operators'
set compile_env: 0
method: PPParser
optional
	"Answer a new parser that parses the receiver, if possible."

	^ PPOptionalParser on: self
%
category: 'parsing'
set compile_env: 0
method: PPParser
parse: anObject
	"Parse anObject with the receiving parser and answer the parse-result or an instance of PPFailure."
	
	^ self parseOn: anObject asPetitStream
%
category: 'parsing'
set compile_env: 0
method: PPParser
parse: anObject onError: aBlock
	"Parse anObject with the receiving parser and answer the parse-result or answer the result of evaluating aBlock. Depending on the number of arguments of the block it is simply evaluated, evaluated with the failure object, or evaluated with the error message and position."
	
	| result |
	result := self parse: anObject.
	result isPetitFailure
		ifFalse: [ ^ result ].
	aBlock numArgs = 0
		ifTrue: [ ^ aBlock value ].
	aBlock numArgs = 1
		ifTrue: [ ^ aBlock value: result ].
	^ aBlock value: result message value: result position
%
category: 'parsing'
set compile_env: 0
method: PPParser
parseOn: aStream
	"Parse aStream with the receiving parser and answer the parse-result or an instance of PPFailure. Override this method in subclasses to specify custom parse behavior. Do not call this method from outside, instead use #parse:."
	
	self subclassResponsibility: #parseOn:
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
plus
	"Answer a new parser that parses the receiver one or more times."

	^ self star setMin: 1
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
plusGreedy: aParser
	"Answer a new parser that parses the receiver one or more times until it reaches aParser. This is a greedy non-blind implementation of the star operator. aParser is not consumed."
	
	^ (self starGreedy: aParser) setMin: 1
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
plusLazy: aParser
	"Answer a new parser that parses the receiver one or more times until it reaches aParser. This is a lazy non-blind implementation of the star operator. aParser is not consumed."
	
	^ (self starLazy: aParser) setMin: 1
%
category: 'copying'
set compile_env: 0
method: PPParser
postCopy
	super postCopy.
	properties := properties copy
%
category: 'printing'
set compile_env: 0
method: PPParser
printNameOn: aStream

	self name isNil
		ifTrue: [self hash printOn: aStream]
		ifFalse: [aStream nextPutAll: self name]
%
category: 'printing'
set compile_env: 0
method: PPParser
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(.
	self printNameOn: aStream.
	aStream nextPut: $)
%
category: 'accessing-properties'
set compile_env: 0
method: PPParser
propertyAt: aKey
	"Answer the property value associated with aKey."
	
	^ self propertyAt: aKey ifAbsent: [ self error: 'Property not found' ]
%
category: 'accessing-properties'
set compile_env: 0
method: PPParser
propertyAt: aKey ifAbsent: aBlock
	"Answer the property value associated with aKey or, if aKey isn't found, answer the result of evaluating aBlock."
	
	^ properties isNil
		ifTrue: [ aBlock value ]
		ifFalse: [ properties at: aKey ifAbsent: aBlock ]
%
category: 'accessing-properties'
set compile_env: 0
method: PPParser
propertyAt: aKey ifAbsentPut: aBlock
	"Answer the property associated with aKey or, if aKey isn't found store the result of evaluating aBlock as new value."
	
	^ self propertyAt: aKey ifAbsent: [ self propertyAt: aKey put: aBlock value ]
%
category: 'accessing-properties'
set compile_env: 0
method: PPParser
propertyAt: aKey put: anObject
	"Set the property at aKey to be anObject. If aKey is not found, create a new entry for aKey and set is value to anObject. Answer anObject."

	^ (properties ifNil: [ properties := Dictionary new: 1 ])
		at: aKey put: anObject
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
queryToken
  ^ GsQueryTokenParser on: self
%
category: 'accessing-properties'
set compile_env: 0
method: PPParser
removeProperty: aKey
	"Remove the property with aKey. Answer the property or raise an error if aKey isn't found."
	
	^ self removeProperty: aKey ifAbsent: [ self error: 'Property not found' ]
%
category: 'accessing-properties'
set compile_env: 0
method: PPParser
removeProperty: aKey ifAbsent: aBlock
	"Remove the property with aKey. Answer the value or, if aKey isn't found, answer the result of evaluating aBlock."
	
	| answer |
	properties isNil ifTrue: [ ^ aBlock value ].
	answer := properties removeKey: aKey ifAbsent: aBlock.
	properties isEmpty ifTrue: [ properties := nil ].
	^ answer
%
category: 'operators-convenience'
set compile_env: 0
method: PPParser
separatedBy: aParser
	"Answer a new parser that parses the receiver one or more times, separated by aParser."

	^(PPSequenceParser with: self
		with: (PPSequenceParser with: aParser with: self) star) ==> 
				[:nodes |
				| result |
				result := Array new: 2 * (nodes at: 2) size + 1.
				result at: 1 put: nodes first.
				(nodes at: 2) keysAndValuesDo: 
						[:index :pair |
						result
							replaceFrom: 2 * index
							to: 2 * index + 1
							with: pair
							startingAt: 1].
				result]
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
star
	"Answer a new parser that parses the receiver zero or more times. This is a greedy and blind implementation that tries to consume as much input as possible and it does not consider what comes afterwards."

	^ PPPossessiveRepeatingParser on: self
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
starGreedy: aParser
	"Answer a new parser that parses the receiver zero or more times until it reaches aParser. This is a greedy non-blind implementation of the star operator. aParser is not consumed."
	
	^ PPGreedyRepeatingParser on: self limit: aParser
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
starLazy: aParser
	"Answer a new parser that parses the receiver zero or more times until it reaches aParser. This is a lazy non-blind implementation of the star operator. aParser is not consumed."
	
	^ PPLazyRepeatingParser on: self limit: aParser
%
category: 'operators-repeating'
set compile_env: 0
method: PPParser
times: anInteger
	"Answer a new parser that parses the receiver exactly anInteger times."
	
	^ self min: anInteger max: anInteger
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
token
	"Answer a new parser that transforms the input to a token."
	
	^ PPTokenParser on: self
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
token: aTokenClass
	"Answer a new parser that transforms the input to a token of class aTokenClass."
	
	^ self token tokenClass: aTokenClass
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
trim
	"Answer a new parser that consumes spaces before and after the receiving parser."
	
	^ self trimSpaces
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
trim: aParser
	"Answer a new parser that consumes and ignores aParser repeatedly before and after the receiving parser."
	
	^ PPTrimmingParser on: self trimmer: aParser
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
trimBlanks
	"Answer a new parser that consumes blanks before and after the receiving parser."
	
	^ self trim: #blank asParser
%
category: 'operators-mapping'
set compile_env: 0
method: PPParser
trimSpaces
	"Answer a new parser that consumes spaces before and after the receiving parser."
	
	^ self trim: #space asParser
%
category: 'operators'
set compile_env: 0
method: PPParser
wrapped
	"Answer a new parser that is simply wrapped."
	
	^ PPDelegateParser on: self
%
category: 'operators'
set compile_env: 0
method: PPParser
| aParser
	"Answer a new parser that either parses the receiver or aParser. Fail if both pass or fail (exclusive choice, unordered choice)."

	^(self not , aParser) / (aParser not , self) ==> [:nodes | nodes at: 2]
%
expectValue %Boolean
doit
PPParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPPluggableParser
removeallmethods PPPluggableParser
removeallclassmethods PPPluggableParser
! ------------------- Class methods for PPPluggableParser
category: 'instance creation'
set compile_env: 0
classmethod: PPPluggableParser
on: aBlock
	^ self new initializeOn: aBlock
%
! ------------------- Instance methods for PPPluggableParser
category: 'accessing'
set compile_env: 0
method: PPPluggableParser
block
	"Answer the pluggable block."

	^ block
%
category: 'initialization'
set compile_env: 0
method: PPPluggableParser
initializeOn: aBlock
	block := aBlock
%
category: 'parsing'
set compile_env: 0
method: PPPluggableParser
parseOn: aStream
	| position result |
	position := aStream position.
	result := block value: aStream.
	result isPetitFailure
		ifTrue: [ aStream position: position ].
	^ result
%
expectValue %Boolean
doit
PPPluggableParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPPossessiveRepeatingParser
removeallmethods PPPossessiveRepeatingParser
removeallclassmethods PPPossessiveRepeatingParser
! ------------------- Class methods for PPPossessiveRepeatingParser
! ------------------- Instance methods for PPPossessiveRepeatingParser
category: 'parsing'
set compile_env: 0
method: PPPossessiveRepeatingParser
parseOn: aStream
	| start element elements |
	start := aStream position.
	elements := OrderedCollection new.
	[ elements size < min ] whileTrue: [
		(element := parser parseOn: aStream) isPetitFailure ifTrue: [
			aStream position: start.
			^ element ].
		elements addLast: element ].
	[ elements size < max ] whileTrue: [
	 	(element := parser parseOn: aStream) isPetitFailure
			ifTrue: [ ^ elements asArray ].
		elements addLast: element ].
	^ elements asArray
%
expectValue %Boolean
doit
PPPossessiveRepeatingParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPPredicateObjectParser
removeallmethods PPPredicateObjectParser
removeallclassmethods PPPredicateObjectParser
! ------------------- Class methods for PPPredicateObjectParser
category: 'factory-objects'
set compile_env: 0
classmethod: PPPredicateObjectParser
any
	^ self
		on: [ :each | true ] message: 'input expected'
		negated: [ :each | false ] message: 'no input expected'
%
category: 'factory-objects'
set compile_env: 0
classmethod: PPPredicateObjectParser
anyExceptAnyOf: aCollection
	^ self
		on: [ :each | (aCollection includes: each) not ] message: 'any except ' , aCollection printString , ' expected'
		negated: [ :each | aCollection includes: each ] message: aCollection printString ,  ' not expected'
%
category: 'factory-objects'
set compile_env: 0
classmethod: PPPredicateObjectParser
anyOf: aCollection
	^ self
		on: [ :each | aCollection includes: each ] message: 'any of ' , aCollection printString , ' expected'
		negated: [ :each | (aCollection includes: each) not ] message: 'none of ' , aCollection printString ,  'expected'
%
category: 'factory-objects'
set compile_env: 0
classmethod: PPPredicateObjectParser
between: min and: max
	^ self
		on: [ :each | each >= min and: [ each <= max ] ] message: min printString , '..' , max printString , ' expected'
		negated: [ :each | each < min or: [ each > max ] ] message: min printString , '..' , max printString , ' not expected'
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
blank
	^ self chars: (String with: Character space with: Character tab) message: 'blank expected'
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
char: aCharacter
	^ self expect: aCharacter message: (String with: $" with: aCharacter with: $") , ' expected'
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
char: aCharacter message: aString
	^ self expect: aCharacter message: aString
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
chars: aCollection message: aString
	^ self on: (PPCharSetPredicate on: [ :char | aCollection includes: char ]) message: aString
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
cr
	^ self char: Character cr message: 'carriage return expected'
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
digit
	^ self on: (PPCharSetPredicate on: [ :char | char isDigit ]) message: 'digit expected'
%
category: 'factory-objects'
set compile_env: 0
classmethod: PPPredicateObjectParser
expect: anObject
	^ self expect: anObject message: anObject printString , ' expected'
%
category: 'factory-objects'
set compile_env: 0
classmethod: PPPredicateObjectParser
expect: anObject message: aString
	^ self 
		on: [ :each | each = anObject ] message: aString
		negated: [ :each | each ~= anObject ] message: 'no ' , aString
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
hex
	^ self 
		on: (PPCharSetPredicate on: [ :char | 
			(char between: $0 and: $9) 
				or: [ (char between: $a and: $f) 
				or: [ (char between: $A and: $F) ] ] ])
		message: 'hex digit expected'
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
letter
	^ self on: (PPCharSetPredicate on: [ :char | char isLetter ]) message: 'letter expected'
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
lf
	^ self char: Character lf
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
lowercase
	^ self on: (PPCharSetPredicate on: [ :char | char isLowercase ]) message: 'lowercase letter expected'
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
newline
	^ self chars: (String with: Character cr with: Character lf) message: 'newline expected'
%
category: 'instance creation'
set compile_env: 0
classmethod: PPPredicateObjectParser
on: aBlock message: aString
	^ self on: aBlock message: aString negated: [ :each | (aBlock value: each) not ] message: 'no ' , aString
%
category: 'instance creation'
set compile_env: 0
classmethod: PPPredicateObjectParser
on: aBlock message: aString negated: aNegatedBlock message: aNegatedString
	^ self new initializeOn: aBlock message: aString negated: aNegatedBlock message: aNegatedString
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
punctuation
	^ self chars: '.,"''?!;:#$%&()*+-/<>=@[]\^_{}|~' message: 'punctuation expected'
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
space
	^ self on: (PPCharSetPredicate on: [ :char | char isSeparator ]) message: 'separator expected'
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
tab
	^ self char: Character tab message: 'tab expected'
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
uppercase
	^ self on: (PPCharSetPredicate on: [ :char | char isUppercase ]) message: 'uppercase letter expected'
%
category: 'factory-chars'
set compile_env: 0
classmethod: PPPredicateObjectParser
word
	^ self on: (PPCharSetPredicate on: [ :char | char isAlphaNumeric ]) message: 'letter or digit expected'
%
! ------------------- Instance methods for PPPredicateObjectParser
category: 'initialization'
set compile_env: 0
method: PPPredicateObjectParser
initializeOn: aBlock message: aString negated: aNegatedBlock message: aNegatedString
	predicate := aBlock.
	predicateMessage := aString.
	negated := aNegatedBlock.
	negatedMessage := aNegatedString
%
category: 'operators'
set compile_env: 0
method: PPPredicateObjectParser
negate
	"Answer a parser that is the negation of the receiving predicate parser."
	
	^ self class 
		on: negated message: negatedMessage 
		negated: predicate message: predicateMessage
%
category: 'parsing'
set compile_env: 0
method: PPPredicateObjectParser
parseOn: aStream
	^ (aStream atEnd not and: [ predicate value: aStream uncheckedPeek ])
		ifFalse: [ PPFailure message: predicateMessage at: aStream position ]
		ifTrue: [ aStream next ]
%
expectValue %Boolean
doit
PPPredicateObjectParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPPredicateParser
removeallmethods PPPredicateParser
removeallclassmethods PPPredicateParser
! ------------------- Class methods for PPPredicateParser
! ------------------- Instance methods for PPPredicateParser
category: 'accessing'
set compile_env: 0
method: PPPredicateParser
block
	"Answer the predicate block of the receiver."
	
	^ predicate
%
category: 'accessing'
set compile_env: 0
method: PPPredicateParser
message
	"Answer the failure message."
	
	^ predicateMessage
%
category: 'printing'
set compile_env: 0
method: PPPredicateParser
printNameOn: aStream

	super printNameOn: aStream.
	aStream nextPutAll: ', '.
	predicateMessage printOn: aStream
%
expectValue %Boolean
doit
PPPredicateParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPPredicateSequenceParser
removeallmethods PPPredicateSequenceParser
removeallclassmethods PPPredicateSequenceParser
! ------------------- Class methods for PPPredicateSequenceParser
category: 'instance creation'
set compile_env: 0
classmethod: PPPredicateSequenceParser
on: aBlock message: aString negated: aNegatedBlock message: aNegatedString size: anInteger 
	^ self new initializeOn: aBlock message: aString negated: aNegatedBlock message: aNegatedString size: anInteger
%
category: 'instance creation'
set compile_env: 0
classmethod: PPPredicateSequenceParser
on: aBlock message: aString size: anInteger
	^ self on: aBlock message: aString negated: [ :each | (aBlock value: each) not ] message: 'no ' , aString size: anInteger 
%
! ------------------- Instance methods for PPPredicateSequenceParser
category: 'initialization'
set compile_env: 0
method: PPPredicateSequenceParser
initializeOn: aBlock message: aString negated: aNegatedBlock message: aNegatedString size: anInteger
	predicate := aBlock.
	predicateMessage := aString.
	negated := aNegatedBlock.
	negatedMessage := aNegatedString.
	size := anInteger 
%
category: 'operators'
set compile_env: 0
method: PPPredicateSequenceParser
negate
	"Answer a parser that is the negation of the receiving predicate parser."
	
	^ self class 
		on: negated message: negatedMessage
		negated: predicate message: predicateMessage
		size: size
%
category: 'parsing'
set compile_env: 0
method: PPPredicateSequenceParser
parseOn: aStream
	| position result |
	position := aStream position.
	result := aStream next: size.
	(result size = size and: [ predicate value: result ])
		ifTrue: [ ^ result ].
	aStream position: position.
	^ PPFailure message: predicateMessage at: aStream position
%
category: 'accessing'
set compile_env: 0
method: PPPredicateSequenceParser
size
	"Answer the sequence size of the receiver."

	^ size
%
expectValue %Boolean
doit
PPPredicateSequenceParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPRepeatingParser
removeallmethods PPRepeatingParser
removeallclassmethods PPRepeatingParser
! ------------------- Class methods for PPRepeatingParser
! ------------------- Instance methods for PPRepeatingParser
category: 'initialization'
set compile_env: 0
method: PPRepeatingParser
initialize

	super initialize.
	self
		setMin: 0;
		setMax: SmallInteger maximumValue
%
category: 'accessing'
set compile_env: 0
method: PPRepeatingParser
max
	"Answer the maximum number of repetitions."

	^ max
%
category: 'accessing'
set compile_env: 0
method: PPRepeatingParser
min
	"Answer the minimum number of repetitions."
	
	^ min
%
category: 'printing'
set compile_env: 0
method: PPRepeatingParser
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: ' ['.
	min printOn: aStream.
	aStream
		nextPutAll: ', ';
		nextPutAll: (max = SmallInteger maximumValue
					ifTrue: ['*']
					ifFalse: [max printString]);
		nextPut: $]
%
category: 'initialization'
set compile_env: 0
method: PPRepeatingParser
setMax: anInteger
	max := anInteger
%
category: 'initialization'
set compile_env: 0
method: PPRepeatingParser
setMin: anInteger
	min := anInteger
%
expectValue %Boolean
doit
PPRepeatingParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPSequenceParser
removeallmethods PPSequenceParser
removeallclassmethods PPSequenceParser
! ------------------- Class methods for PPSequenceParser
! ------------------- Instance methods for PPSequenceParser
category: 'operators'
set compile_env: 0
method: PPSequenceParser
, aRule
	^ self copyWith: aRule
%
category: 'operators-mapping'
set compile_env: 0
method: PPSequenceParser
map: aBlock
	^ aBlock numArgs = self children size
		ifTrue: [ self ==> [ :nodes | aBlock valueWithArguments: nodes ] ]
		ifFalse: [ self error: aBlock numArgs asString , ' arguments expected.' ]
%
category: 'parsing'
set compile_env: 0
method: PPSequenceParser
parseOn: aStream
	"This is optimized code that avoids unnecessary block activations, do not change."
	
	| start elements element |
	start := aStream position.
	elements := Array new: parsers size.
	1 to: parsers size do: [ :index |
		element := (parsers at: index) 
			parseOn: aStream.
		element isPetitFailure ifTrue: [
			aStream position: start.
			^ element ].
		elements at: index put: element ].
	^ elements
%
category: 'operators-mapping'
set compile_env: 0
method: PPSequenceParser
permutation: anArrayOfIntegers
	"Answer a permutation of the receivers sequence."

	anArrayOfIntegers do: 
			[:index |
			(index _isInteger and: [index between: 1 and: parsers size])
				ifFalse: [self error: 'Invalid permutation index: ' , index printString]].
	^self ==> [:nodes | anArrayOfIntegers collect: [:index | nodes at: index]]
%
expectValue %Boolean
doit
PPSequenceParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPToken
removeallmethods PPToken
removeallclassmethods PPToken
! ------------------- Class methods for PPToken
category: 'instance creation'
set compile_env: 0
classmethod: PPToken
new
	self error: 'Token can only be created using a dedicated constructor.'
%
category: 'instance creation'
set compile_env: 0
classmethod: PPToken
on: aSequenceableCollection
	^ self on: aSequenceableCollection start: 1 stop: aSequenceableCollection size
%
category: 'instance creation'
set compile_env: 0
classmethod: PPToken
on: aSequenceableCollection start: aStartInteger stop: aStopInteger
	^ self basicNew 
		initializeOn: aSequenceableCollection
		start: aStartInteger stop: aStopInteger
%
! ------------------- Instance methods for PPToken
category: 'comparing'
set compile_env: 0
method: PPToken
= anObject
	^ self class = anObject class and: [ self value = anObject value ]
%
category: 'accessing'
set compile_env: 0
method: PPToken
collection
	"Answer the underlying collection of this token."

	^ collection
%
method: PPToken
newLineParser
	^newLineParser
%
method: PPToken
newLineParser: aParser
	newLineParser := aParser
%
category: 'querying'
set compile_env: 0
method: PPToken
column
	"Answer the column number of this token in the underlying collection."
	
	| position |
	position := 0.
	(self newLineParser , [ :stream |
		start <= stream position
			ifTrue: [ ^ start - position ].
		position := stream position ] asParser
		/ #any asParser) star
			parse: collection.
	 ^ start - position
%
category: 'copying'
set compile_env: 0
method: PPToken
copyFrom: aStartInteger to: aStopInteger
	^ self class on: collection start: start + aStartInteger - 1 stop: stop + aStopInteger - 3
%
category: 'comparing'
set compile_env: 0
method: PPToken
hash
	^ self value hash
%
category: 'initialization'
set compile_env: 0
method: PPToken
initializeOn: aSequenceableCollection start: aStartInteger stop: aStopInteger
    
	newLineParser := (Character lf asParser) / (Character cr asParser , Character lf asParser optional).
	collection := aSequenceableCollection.
	start := aStartInteger.
	stop := aStopInteger
%
category: 'querying'
set compile_env: 0
method: PPToken
line
	"Answer the line number of this token in the underlying collection."
	
	| line |
	line := 1.
	(self newLineParser , [ :stream |
		start <= stream position
			ifTrue: [ ^ line ].
		line := line + 1 ] asParser
		/ #any asParser) star
			parse: collection.
	^ line
%
category: 'printing'
set compile_env: 0
method: PPToken
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(; nextPutAll: self value; nextPut: $)
%
category: 'accessing'
set compile_env: 0
method: PPToken
size
	"Answer the size of this token."

	^ stop - start + 1
%
category: 'accessing'
set compile_env: 0
method: PPToken
start
	"Answer the start position of this token in the underlying collection."

	^ start
%
category: 'accessing'
set compile_env: 0
method: PPToken
stop
	"Answer the stop position of this token in the underlying collection."
	
	^ stop
%
category: 'accessing'
set compile_env: 0
method: PPToken
value
	"Answer the contents of this token."

	^ collection copyFrom: start to: stop
%
expectValue %Boolean
doit
PPToken category: 'PetitParser-Core'.
true
%

! Remove existing behavior from PPTokenParser
removeallmethods PPTokenParser
removeallclassmethods PPTokenParser
! ------------------- Class methods for PPTokenParser
! ------------------- Instance methods for PPTokenParser
category: 'private'
set compile_env: 0
method: PPTokenParser
create: aCollection start: aStartInteger stop: aStopInteger
	^ self tokenClass on: aCollection start: aStartInteger stop: aStopInteger
%
category: 'private'
set compile_env: 0
method: PPTokenParser
defaultTokenClass
	^ PPToken
%
category: 'initialization'
set compile_env: 0
method: PPTokenParser
initialize
	tokenClass := self defaultTokenClass
	
%
category: 'accessing'
set compile_env: 0
method: PPTokenParser
tokenClass
	^ tokenClass
%
category: 'accessing'
set compile_env: 0
method: PPTokenParser
tokenClass: aTokenClass
	tokenClass := aTokenClass
%
expectValue %Boolean
doit
PPTokenParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPTrimmingParser
removeallmethods PPTrimmingParser
removeallclassmethods PPTrimmingParser
! ------------------- Class methods for PPTrimmingParser
category: 'instance creation'
set compile_env: 0
classmethod: PPTrimmingParser
on: aParser trimmer: aTrimParser
	^ self new
		setParser: aParser;
		setTrimmer: aTrimParser;
		yourself
%
! ------------------- Instance methods for PPTrimmingParser
category: 'parsing'
set compile_env: 0
method: PPTrimmingParser
parseOn: aStream
	| position element |
	position := aStream position.
	[ (trimmer parseOn: aStream) isPetitFailure ]
		whileFalse.
	element := parser parseOn: aStream.
	element isPetitFailure ifTrue: [
		aStream position: position.
		^ element ].
	[ (trimmer parseOn: aStream) isPetitFailure ]
		whileFalse.
	^ element
%
category: 'initialization'
set compile_env: 0
method: PPTrimmingParser
setTrimmer: aParser
	trimmer := aParser
%
expectValue %Boolean
doit
PPTrimmingParser category: 'PetitParser-Parsers'.
true
%

! Remove existing behavior from PPUnresolvedParser
removeallmethods PPUnresolvedParser
removeallclassmethods PPUnresolvedParser
! ------------------- Class methods for PPUnresolvedParser
! ------------------- Instance methods for PPUnresolvedParser
category: 'testing'
set compile_env: 0
method: PPUnresolvedParser
isUnresolved
	^ true
%
category: 'parsing'
set compile_env: 0
method: PPUnresolvedParser
parseOn: aStream
	self error: self printString , ' need to be resolved before execution.'
%
expectValue %Boolean
doit
PPUnresolvedParser category: 'PetitParser-Tools'.
true
%

! Remove existing behavior from PPWrappingParser
removeallmethods PPWrappingParser
removeallclassmethods PPWrappingParser
! ------------------- Class methods for PPWrappingParser
! ------------------- Instance methods for PPWrappingParser
category: 'parsing'
set compile_env: 0
method: PPWrappingParser
parseOn: aStream
	^ block value: aStream value: [ parser parseOn: aStream ]
%
expectValue %Boolean
doit
PPWrappingParser category: 'PetitParser-Parsers'.
true
%
! ------------------- Class initializers 
