! file  sortedcollection2.gs

category: 'Repository Conversion'
classmethod: SortedCollection
convertAll: anArray usingCache: dict hiddenSetForErrors: hsIdx

"Convert an Array of SortedCollections by recompiling their sort blocks from
 SimpleBlocks to ExecBlocks.  anArray must contain only SortedCollections.
 dict is used as a cache to canonicalize sort blocks with the same source 
 string.

 Caller must handle and catch AlmostOutOfMemory errors and handle commits
 and aborts."

| sb eb |
sb := Globals at:#SimpleBlock ifAbsent:[ (Globals at:#ObsoleteClasses) at: #SimpleBlock ] .
eb := ExecBlock .
1 to: anArray size do:[:n| | sys sc sortBlock val cls errorBlock |
  sys := System .
  errorBlock := [:_sys :_sc :_hsIdx| 
                     _sys _add: _sc to: _hsIdx .
                     _sys _sessionCacheStatAt: 1 put: (sys _hiddenSetSize: _hsIdx) ].
  sc := anArray at: n .
  sortBlock := sc sortBlock .
  cls := sortBlock class .
  cls == sb "Block from 2.x.  Attempt to convert"
    ifTrue:[ | str |
      str := sortBlock _sourceString .
      val := dict at: str otherwise: nil .
      val ifNil:[ "A miss on the dictionary, compile it now"
            [ val := str evaluate ]  onSynchronous: CompileError
                                     do:[:ex | errorBlock value: sys value: sc value: hsIdx.
                                               val := nil. 
                                               ex resume ] .
            val   "Now see if the recompile was successful"
              "Add to error list.  Probably already done."
              ifNil:[ errorBlock value: sys value: sc value: hsIdx ] 
              ifNotNil:[ dict at: str put: val. "Add to cache"
                         sc _sortBlock: val]. "Store in instance"
      ] ifNotNil:[ sc _sortBlock: val ]. "A hit on the dictionary, store the block"
  ] ifFalse:[ "Not a SimpleBlock"
      cls == eb "Ignore if already converted"
        ifFalse:[ errorBlock value: sys value: sc value: hsIdx ]. "Probably complex or complexVC block."
  ].
].
%

